@@ -8,20 +8,35 @@ from pipelime.piper import PipelimeCommand, PiperPortType
 from pipelime.stages import StageInput
 
 
+class OutputTime(pyd.BaseModel):
+    nanosec: int
+
+    def __repr__(self) -> str:
+        return self.__piper_repr__()
+
+    def __piper_repr__(self) -> str:
+        from pipelime.cli.utils import time_to_str
+
+        return time_to_str(self.nanosec)
+
+
+class OutputStageTime(pyd.BaseModel):
+    stages: t.Mapping[str, OutputTime]
+
+    def __repr__(self) -> str:
+        return self.__piper_repr__()
+
+    def __piper_repr__(self) -> str:
+        return (
+            "{\n  "
+            + "\n  ".join(f"{k}: {repr(v)}" for k, v in self.stages.items())
+            + "\n}"
+        )
+
+
 class TimeItCommand(PipelimeCommand, title="timeit"):
     """Measures the average time to get a sample from a sequence."""
 
-    class OutputTime(pyd.BaseModel):
-        nanosec: int
-
-        def __repr__(self) -> str:
-            return self.__piper_repr__()
-
-        def __piper_repr__(self) -> str:
-            from pipelime.cli.utils import time_to_str
-
-            return time_to_str(self.nanosec)
-
     input: t.Optional[
         pl_interfaces.InputDatasetInterface
     ] = pl_interfaces.InputDatasetInterface.pyd_field(
@@ -129,9 +144,126 @@ class TimeItCommand(PipelimeCommand, title="timeit"):
             num_samples += available_samples
             elapsed_times.append(end - start)
 
-        self.average_time = TimeItCommand.OutputTime(
-            nanosec=int(sum(elapsed_times) // num_samples)
+        self.average_time = OutputTime(nanosec=int(sum(elapsed_times) // num_samples))
+
+
+class StageTimingCommand(PipelimeCommand, title="stage-time"):
+    """Measures the average time to get a sample through some stages."""
+
+    stages: t.Union[StageInput, t.Sequence[StageInput]] = pyd.Field(
+        ...,
+        alias="s",
+        description=(
+            "One or more stages to apply. Can be a stage name/class_path "
+            "(with no arguments) or a dictionary with the stage name/class_path as key "
+            "and the arguments passed by keywords."
+        ),
+    )
+
+    input: pl_interfaces.InputDatasetInterface = (
+        pl_interfaces.InputDatasetInterface.pyd_field(
+            alias="i", piper_port=PiperPortType.INPUT
         )
+    )
+
+    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
+        alias="g"
+    )
+
+    skip_first: pyd.NonNegativeInt = pyd.Field(
+        1,
+        alias="s",
+        description="Skip the first n samples in EACH EXECUTION PROCESS, then start the timer.",
+    )
+    max_samples: t.Optional[pyd.PositiveInt] = pyd.Field(
+        None,
+        alias="m",
+        description="Grab at most `max_samples` from THE WHOLE SEQUENCE and take the average time.",
+    )
+    repeat: pyd.PositiveInt = pyd.Field(
+        1, alias="r", description="Repeat the measurement `repeat` times."
+    )
+    process: bool = pyd.Field(
+        False,
+        alias="p",
+        description=(
+            "Measure process time instead of using a performance counter clock."
+        ),
+    )
+
+    average_time: t.Optional[OutputStageTime] = pyd.Field(
+        None,
+        description="The average time to get a sample through the stages.",
+        exclude=True,
+        repr=False,
+        piper_port=PiperPortType.OUTPUT,
+    )
+
+    def run(self):
+        from pipelime.stages import StageTimer, StageCompose
+
+        # create unique names
+        stages = self.stages if isinstance(self.stages, t.Sequence) else [self.stages]
+        names = []
+        for st in stages:
+            stage_cls = st.__root__.__class__
+            stage_name = (
+                stage_cls.__config__.title
+                if stage_cls.__config__.title
+                else stage_cls.__name__
+            )
+            names.append(stage_name)
+        for i, n in enumerate(names[::-1]):
+            names[-i - 1] = f"{n}-{names.count(n)}"
+
+        class _AverageTime:
+            def __init__(self):
+                self.timings = {}
+                self.timings_key = "~timings"
+
+            def __call__(self, x):
+                if self.timings_key in x:
+                    tt = x[self.timings_key]()
+                    for k, v in tt.items():
+                        ctime, celem = self.timings.setdefault(k, (0, 0))
+                        self.timings[k] = (ctime + v, celem + 1)
+
+            def to_output(self):
+                return OutputStageTime(
+                    stages={
+                        k: OutputTime(nanosec=int(tt // n))
+                        for k, (tt, n) in self.timings.items()
+                    }
+                )
+
+        avg_time = _AverageTime()
+
+        for r in range(self.repeat):
+            seq = self.input.create_reader()
+            if self.max_samples:
+                seq = seq[: self.max_samples]
+            seq = seq.map(
+                stage=StageCompose(
+                    stages=[
+                        StageTimer(
+                            stage=st,  # type: ignore
+                            skip_first=self.skip_first,
+                            time_key_path=f"{avg_time.timings_key}.{n}",
+                            process=self.process,
+                        )
+                        for st, n in zip(stages, names)
+                    ]
+                )
+            )
+            self.grabber.grab_all(
+                seq,
+                keep_order=False,
+                parent_cmd=self,
+                sample_fn=avg_time,
+                track_message=f"Timing data ({len(seq)} samples)",
+            )
+
+        self.average_time = avg_time.to_output()
 
 
 class PipeCommand(PipelimeCommand, title="pipe"):
@@ -624,6 +756,61 @@ class MapCommand(PipelimeCommand, title="map"):
         )
 
 
+class MapIfCommand(PipelimeCommand, title="map-if"):
+    """Apply a stage on a dataset if a condition is met."""
+
+    stage: StageInput = pyd.Field(
+        ...,
+        alias="s",
+        description=(
+            "A stage to apply. Can be a stage name/class_path (with no arguments) or "
+            "a dictionary with the stage name/class_path as key and the arguments "
+            "passed by keywords."
+        ),
+    )
+
+    input: pl_interfaces.InputDatasetInterface = (
+        pl_interfaces.InputDatasetInterface.pyd_field(
+            alias="i", piper_port=PiperPortType.INPUT
+        )
+    )
+
+    output: pl_interfaces.OutputDatasetInterface = (
+        pl_interfaces.OutputDatasetInterface.pyd_field(
+            alias="o", piper_port=PiperPortType.OUTPUT
+        )
+    )
+
+    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
+        alias="g"
+    )
+
+    condition: pl_types.CallableDef = pyd.Field(
+        ...,
+        alias="c",
+        description=(
+            "A `class.path.func`, `file.py:func`, `lambda...` or `func:::def func...` "
+            "of a callable that returns True if the sample should be mapped through "
+            "the stage. Accepted signatures:\n"
+            "  `() -> bool`\n"
+            "  `(index: int) -> bool`\n"
+            "  `(index: int, sample: Sample) -> bool`\n"
+            "  `(index: int, sample: Sample, source: SamplesSequence) -> bool`."
+        ),
+    )
+
+    def run(self):
+        seq = self.input.create_reader()
+        seq = seq.map_if(stage=self.stage, condition=self.condition)
+        self.grabber.grab_all(
+            self.output.append_writer(seq),
+            grab_context_manager=self.output.serialization_cm(),
+            keep_order=False,
+            parent_cmd=self,
+            track_message=f"Mapping data ({len(seq)} samples)",
+        )
+
+
 class SortCommand(PipelimeCommand, title="sort"):
     """Sort a dataset by metadata values or according to a custom sorting function."""
 
@@ -640,7 +827,8 @@ class SortCommand(PipelimeCommand, title="sort"):
         None,
         alias="f",
         description=(
-            "A class path to a callable `(Sample) -> Any` to be used as key-function. "
+            "A `class.path.func`, `file.py:func`, `lambda...` or `func:::def func...` "
+            "of a callable `(Sample) -> Any` to be used as key-function. "
             "Use `functools.cmp_to_key` to convert a compare function, "
             "ie, accepting two arguments, to a key function."
         ),
@@ -698,8 +886,8 @@ class FilterCommand(PipelimeCommand, title="filter"):
         None,
         alias="f",
         description=(
-            "A `class.path.to.func` (or `file.py:func`) to "
-            "a callable `(Sample) -> bool` returning True for any valid sample."
+            "A `class.path.func`, `file.py:func`, `lambda...` or `func:::def func...` "
+            "of a callable `(Sample) -> bool` returning True for any valid sample."
         ),
     )
 
@@ -825,8 +1013,8 @@ class SetMetadataCommand(FilterCommand, title="set-meta"):
         None,
         alias="f",
         description=(
-            "A `class.path.to.func` (or `file.py:func`) to "
-            "a callable returning True for any valid sample.\n"
+            "A `class.path.func`, `file.py:func`, `lambda...` or `func:::def func...` "
+            "of a callable returning True for any valid sample.\n"
             "Accepted signatures:\n"
             "  `() -> bool`\n"
             "  `(index: int) -> bool`\n"
